/*
 * ProviewR   Open Source Process Control.
 * Copyright (C) 2005-2019 SSAB EMEA AB.
 *
 * This file is part of ProviewR.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ProviewR. If not, see <http://www.gnu.org/licenses/>
 *
 * Linking ProviewR statically or dynamically with other modules is
 * making a combined work based on ProviewR. Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * In addition, as a special exception, the copyright holders of
 * ProviewR give you permission to, from the build function in the
 * ProviewR Configurator, combine ProviewR with modules generated by the
 * ProviewR PLC Editor to a PLC program, regardless of the license
 * terms of these modules. You may copy and distribute the resulting
 * combined work under the terms of your choice, provided that every
 * copy of the combined work is accompanied by a complete copy of
 * the source code of ProviewR (the version used to produce the
 * combined work), being distributed under the terms of the GNU
 * General Public License plus this exception.
 */

#include <ctype.h>
#include <string.h>
#include <math.h>

#include <iostream>

#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#if defined PWRE_CONF_RSVG
#include <rsvg.h>
#endif

#include "co_string.h"
#include "glow_browctx.h"
#include "glow_colpalctx.h"
#include "glow_curvectx.h"
#include "glow_customcolors_gtk.h"
#include "glow_keyboardctx.h"
#include "glow_msg.h"

using namespace std;

#define DRAW_PRESS_PIX 9

typedef struct {
  GtkWidget w;
  int x;
  int y;
  int width;
  int height;
  GlowArrayElem* node;
  int number;
  GlowCtx* ctx;
} draw_sAnnotData;

typedef struct {
  GlowCtx* ctx;
  void (*callback_func)(GlowCtx* ctx);
  guint timer_id;
} draw_sTimerCb;

typedef struct {
  Pixmap pixmap[DRAW_PIXMAP_SIZE];
} draw_sPixmap;

typedef struct {
  int red;
  int green;
  int blue;
} draw_sColor;

static char font_names[glow_eFont__][glow_eDrawFont__][DRAW_FONT_SIZE][80] = {
  { { "-*-Helvetica-Medium-R-Normal--8-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--10-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--12-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--14-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--14-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--18-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--18-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--18-*-*-*-P-*-ISO8859-1",
        "-*-Helvetica-Bold-R-Normal--24-*-*-*-P-*-ISO8859-1" },
      { "-*-Helvetica-Medium-R-Normal--8-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--10-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--12-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--14-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--14-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--18-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--18-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--18-*-*-*-P-*-ISO8859-1",
          "-*-Helvetica-Medium-R-Normal--24-*-*-*-P-*-ISO8859-1" } },
  { { "-*-times-Medium-R-Normal--8-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--10-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--12-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--14-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--14-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--18-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--18-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--18-*-*-*-*-*-ISO8859-1",
        "-*-times-Bold-R-Normal--24-*-*-*-*-*-ISO8859-1" },
      { "-*-times-Medium-R-Normal--8-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--10-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--12-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--14-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--14-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--18-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--18-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--18-*-*-*-*-*-ISO8859-1",
          "-*-times-Medium-R-Normal--24-*-*-*-*-*-ISO8859-1" } },
  { { "-*-new century schoolbook-Medium-R-Normal--8-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--10-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--12-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--14-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--14-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--18-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--18-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--18-*-*-*-*-*-ISO8859-1",
        "-*-new century schoolbook-Bold-R-Normal--24-*-*-*-*-*-ISO8859-1" },
      { "-*-new century schoolbook-Medium-R-Normal--8-*-*-*-*-*-ISO8859-1",
          "-*-new century schoolbook-Medium-R-Normal--10-*-*-*-*-*-ISO8859-1",
          "-*-new century schoolbook-Medium-R-Normal--12-*-*-*-*-*-ISO8859-1",
          "-*-new century schoolbook-Medium-R-Normal--14-*-*-*-*-*-ISO8859-1",
          "-*-new century schoolbook-Medium-R-Normal--14-*-*-*-*-*-ISO8859-1",
          "-*-new century schoolbook-Medium-R-Normal--18-*-*-*-*-*-ISO8859-1",
          "-*-new century schoolbook-Medium-R-Normal--18-*-*-*-*-*-ISO8859-1",
          "-*-new century schoolbook-Medium-R-Normal--18-*-*-*-*-*-ISO8859-1",
          "-*-new century "
          "schoolbook-Medium-R-Normal--24-*-*-*-*-*-ISO8859-1" } },
  { { "-*-courier-Medium-R-Normal--8-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--10-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--12-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--14-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--14-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--18-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--18-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--18-*-*-*-m-*-ISO8859-1",
        "-*-courier-Bold-R-Normal--24-*-*-*-m-*-ISO8859-1" },
      { "-*-courier-Medium-R-Normal--8-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--10-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--12-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--14-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--14-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--18-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--18-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--18-*-*-*-m-*-ISO8859-1",
          "-*-courier-Medium-R-Normal--24-*-*-*-m-*-ISO8859-1" } }
};

static GdkEvent last_event;

static GdkColor glow_allocate_named_color(
    GlowDrawGtk* draw_ctx, const char* named_color);
static GdkColor glow_allocate_color(
    GlowDrawGtk* draw_ctx, int rgb_red, int rgb_green, int rgb_blue);
static GdkColor glow_allocate_custom_color(GlowDrawGtk* draw_ctx,
    glow_eDrawType drawtype, int rgb_red, int rgb_green, int rgb_blue);
static void event_timer(GlowDrawGtk* ctx, int time_ms);
static void cancel_event_timer(GlowDrawGtk* ctx);
static gboolean event_timer_cb(void* ctx);
static int glow_read_color_file(
    const char* filename, draw_sColor** color_array, int* size);

static GdkGC* get_gc(GlowDrawGtk* draw_ctx, int i, int j)
{
  int crea = 0;

  if (GlowColor::is_custom((glow_eDrawType)i)) {
    if (!draw_ctx->get_customcolors())
      return draw_ctx->gcs[0][0];
    if (!draw_ctx->get_customcolors()->gcs[i - glow_eDrawType_CustomColor1][j])
      crea = 1;
  } else if (!draw_ctx->gcs[i][j])
    crea = 1;

  if (crea) {
    GdkGCValues xgcv;
    double r, g, b;

    GlowColor::rgb_color(i, &r, &g, &b, draw_ctx->get_customcolors());
    if (GlowColor::is_custom((glow_eDrawType)i))
      xgcv.foreground = glow_allocate_custom_color(draw_ctx, (glow_eDrawType)i,
          int(r * 65535), int(g * 65535), int(b * 65535));
    else
      xgcv.foreground = glow_allocate_color(
          draw_ctx, int(r * 65535), int(g * 65535), int(b * 65535));
    xgcv.background = draw_ctx->background;
    xgcv.cap_style = GDK_CAP_BUTT;
    xgcv.line_width = j + 1;

    if (GlowColor::is_custom((glow_eDrawType)i)) {
      draw_ctx->get_customcolors()->gcs[i - glow_eDrawType_CustomColor1][j]
          = gdk_gc_new_with_values(draw_ctx->m_wind.window, &xgcv,
              (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                  | GDK_GC_LINE_WIDTH | GDK_GC_CAP_STYLE));
    } else
      draw_ctx->gcs[i][j] = gdk_gc_new_with_values(draw_ctx->m_wind.window,
          &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                     | GDK_GC_LINE_WIDTH | GDK_GC_CAP_STYLE));
  }
  if (GlowColor::is_custom((glow_eDrawType)i))
    return draw_ctx->get_customcolors()
        ->gcs[i - glow_eDrawType_CustomColor1][j];
  else
    return draw_ctx->gcs[i][j];
}

static void glow_create_cursor(GlowDrawGtk* draw_ctx)
{
  /* Create some cursors */

  /* Cross cursor */
  draw_ctx->cursors[glow_eDrawCursor_CrossHair]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_CROSSHAIR);

  draw_ctx->cursors[glow_eDrawCursor_DiamondCross]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_DIAMOND_CROSS);

  draw_ctx->cursors[glow_eDrawCursor_Hand]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_HAND2);

  draw_ctx->cursors[glow_eDrawCursor_BottomLeftCorner]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_BOTTOM_LEFT_CORNER);

  draw_ctx->cursors[glow_eDrawCursor_BottomRightCorner]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_BOTTOM_RIGHT_CORNER);

  draw_ctx->cursors[glow_eDrawCursor_BottomSide]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_BOTTOM_SIDE);

  draw_ctx->cursors[glow_eDrawCursor_TopLeftCorner]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_TOP_LEFT_CORNER);

  draw_ctx->cursors[glow_eDrawCursor_TopRightCorner]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_TOP_RIGHT_CORNER);

  draw_ctx->cursors[glow_eDrawCursor_TopSide]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_TOP_SIDE);

  draw_ctx->cursors[glow_eDrawCursor_RightSide]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_RIGHT_SIDE);

  draw_ctx->cursors[glow_eDrawCursor_LeftSide]
      = gdk_cursor_new_for_display(draw_ctx->display, GDK_LEFT_SIDE);
}

static void draw_free_gc(GlowDrawGtk* draw_ctx)
{
  for (int i = 1; i < glow_eDrawCursor__; i++)
    gdk_cursor_unref(draw_ctx->cursors[i]);

  gdk_gc_unref(draw_ctx->gc_inverse);
  for (int i = 0; i < glow_eDrawType_Color__; i++) {
    for (int j = 0; j < DRAW_TYPE_SIZE; j++) {
      if (draw_ctx->gcs[i][j])
        gdk_gc_unref(draw_ctx->gcs[i][j]);
    }
  }

  for (int i = 0; i < glow_eFont__; i++) {
    for (int j = 0; j < glow_eDrawFont__; j++) {
      for (int k = 0; k < DRAW_FONT_SIZE; k++) {
        if (draw_ctx->font[i][j][k])
          gdk_font_unref(draw_ctx->font[i][j][k]);
      }
    }
  }
}

static void glow_create_gc(GlowDrawGtk* draw_ctx, GdkWindow* window)
{
  GdkFont* font;
  GdkGCValues xgcv;
  int i;
  draw_sColor *color_array, *color_p;
  int size, sts;

  memset(&xgcv, 0, sizeof(xgcv));

  /* Inverse gc */
  xgcv.background = glow_allocate_named_color(draw_ctx, "black");
  xgcv.foreground = draw_ctx->background;
  xgcv.cap_style = GDK_CAP_BUTT;
  draw_ctx->gc_inverse = gdk_gc_new_with_values(window, &xgcv,
      (GdkGCValuesMask)(
          GDK_GC_FOREGROUND | GDK_GC_BACKGROUND | GDK_GC_CAP_STYLE));

  /* Black line gc */
  xgcv.foreground = glow_allocate_named_color(draw_ctx, "black");
  xgcv.background = draw_ctx->background;
  xgcv.cap_style = GDK_CAP_BUTT;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    xgcv.line_width = i + 1;

    draw_ctx->gcs[glow_eDrawType_Line][i] = gdk_gc_new_with_values(
        window, &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                           | GDK_GC_LINE_WIDTH | GDK_GC_CAP_STYLE));
  }

  /* Erase line gc */
  xgcv.foreground = draw_ctx->background;
  xgcv.background = draw_ctx->background;
  xgcv.cap_style = GDK_CAP_BUTT;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    xgcv.line_width = i + 1;

    draw_ctx->gcs[glow_eDrawType_LineErase][i] = gdk_gc_new_with_values(
        window, &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                           | GDK_GC_LINE_WIDTH | GDK_GC_CAP_STYLE));
  }

  /* Red line gc */
  xgcv.foreground = glow_allocate_named_color(draw_ctx, "red");
  xgcv.background = draw_ctx->background;
  xgcv.cap_style = GDK_CAP_BUTT;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    xgcv.line_width = i + 1;

    draw_ctx->gcs[glow_eDrawType_LineRed][i] = gdk_gc_new_with_values(
        window, &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                           | GDK_GC_LINE_WIDTH | GDK_GC_CAP_STYLE));
  }

  /* Gray line gc */
  xgcv.foreground = glow_allocate_named_color(draw_ctx, "gray");
  xgcv.background = draw_ctx->background;
  xgcv.cap_style = GDK_CAP_BUTT;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    xgcv.line_width = i + 1;

    draw_ctx->gcs[glow_eDrawType_LineGray][i] = gdk_gc_new_with_values(
        window, &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                           | GDK_GC_LINE_WIDTH | GDK_GC_CAP_STYLE));
  }

  sts = glow_read_color_file(
      "/home/claes/test/ge_colors.dat", &color_array, &size);
  if (ODD(sts)) {
    color_p = color_array;
    for (int j = glow_eDrawType_Color4; j <= glow_eDrawType_Color300; j++) {
      if (j - glow_eDrawType_Color4 >= size)
        break;

      xgcv.foreground = glow_allocate_color(
          draw_ctx, color_p->red, color_p->green, color_p->blue);
      xgcv.background = draw_ctx->background;
      xgcv.cap_style = GDK_CAP_BUTT;
      for (i = 0; i < DRAW_TYPE_SIZE; i++) {
        xgcv.line_width = i + 1;

        draw_ctx->gcs[j][i] = gdk_gc_new_with_values(window, &xgcv,
            (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                | GDK_GC_LINE_WIDTH | GDK_GC_CAP_STYLE));
      }
      color_p++;
    }
    free((char*)color_array);
  }

  /* Dashed line gc */
  xgcv.foreground = glow_allocate_named_color(draw_ctx, "black");
  xgcv.background = draw_ctx->background;
  xgcv.line_style = GDK_LINE_ON_OFF_DASH;
  xgcv.cap_style = GDK_CAP_BUTT;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    xgcv.line_width = i + 1;
    gint8 dashes = 7 + i;

    draw_ctx->gcs[glow_eDrawType_LineDashed][i] = gdk_gc_new_with_values(window,
        &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                   | GDK_GC_LINE_WIDTH | GDK_GC_LINE_STYLE | GDK_GC_CAP_STYLE));
    gdk_gc_set_dashes(
        draw_ctx->gcs[glow_eDrawType_LineDashed][i], 0, &dashes, 1);
  }

  /* Red dashed line gc */
  xgcv.foreground = glow_allocate_named_color(draw_ctx, "red");
  xgcv.background = draw_ctx->background;
  xgcv.line_style = GDK_LINE_ON_OFF_DASH;
  xgcv.cap_style = GDK_CAP_BUTT;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    xgcv.line_width = i + 1;
    gint8 dashes = 7 + i;

    draw_ctx->gcs[glow_eDrawType_LineDashedRed][i]
        = gdk_gc_new_with_values(window, &xgcv,
            (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND
                | GDK_GC_LINE_WIDTH | GDK_GC_LINE_STYLE | GDK_GC_CAP_STYLE));
    gdk_gc_set_dashes(
        draw_ctx->gcs[glow_eDrawType_LineDashedRed][i], 0, &dashes, 1);
  }

  /* Text */
  xgcv.foreground = glow_allocate_named_color(draw_ctx, "black");
  xgcv.background = draw_ctx->background;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    draw_ctx->gcs[glow_eDrawType_TextHelvetica][i]
        = gdk_gc_new_with_values(window, &xgcv,
            (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
  }

  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    draw_ctx->gcs[glow_eDrawType_TextHelveticaBold][i]
        = gdk_gc_new_with_values(window, &xgcv,
            (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
  }

  xgcv.foreground = draw_ctx->background;
  xgcv.background = draw_ctx->background;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    draw_ctx->gcs[glow_eDrawType_TextHelveticaErase][i]
        = gdk_gc_new_with_values(window, &xgcv,
            (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
  }

  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    draw_ctx->gcs[glow_eDrawType_TextHelveticaEraseBold][i]
        = gdk_gc_new_with_values(window, &xgcv,
            (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
  }

  for (i = 0; i < DRAW_FONT_SIZE; i++) {
    font = gdk_font_load(
        font_names[glow_eFont_Helvetica][glow_eDrawFont_HelveticaBold][i]);
    gdk_gc_set_font(draw_ctx->gcs[glow_eDrawType_TextHelveticaBold][i], font);
    gdk_gc_set_font(
        draw_ctx->gcs[glow_eDrawType_TextHelveticaEraseBold][i], font);
    draw_ctx->font[glow_eFont_Helvetica][glow_eDrawFont_HelveticaBold][i]
        = font;
  }
  for (i = 0; i < DRAW_FONT_SIZE; i++) {
    font = gdk_font_load(
        font_names[glow_eFont_Helvetica][glow_eDrawFont_Helvetica][i]);
    gdk_gc_set_font(draw_ctx->gcs[glow_eDrawType_TextHelvetica][i], font);
    gdk_gc_set_font(draw_ctx->gcs[glow_eDrawType_TextHelveticaErase][i], font);
    draw_ctx->font[glow_eFont_Helvetica][glow_eDrawFont_Helvetica][i] = font;
  }
}

static GdkColor glow_allocate_named_color(
    GlowDrawGtk* draw_ctx, const char* named_color)
{
  GdkColor color;

  if (draw_ctx->color_vect_cnt >= (int)(sizeof(draw_ctx->color_vect)
                                      / sizeof(draw_ctx->color_vect[0]))) {
    printf("Glow allocate color error: max number of colors exceeded\n");
    return draw_ctx->color_vect[0];
  }

  if (!gdk_color_parse(named_color, &color))
    gdk_color_parse("black", &color);
  gdk_colormap_alloc_color(draw_ctx->colormap, &color, FALSE, TRUE);

  draw_ctx->color_vect[draw_ctx->color_vect_cnt] = color;
  draw_ctx->color_vect_cnt++;

  return color;
}

static GdkColor glow_allocate_color(
    GlowDrawGtk* draw_ctx, int rgb_red, int rgb_green, int rgb_blue)
{
  GdkColor color;

  if (draw_ctx->color_vect_cnt >= (int)(sizeof(draw_ctx->color_vect)
                                      / sizeof(draw_ctx->color_vect[0]))) {
    printf("Glow allocate color error: max number of colors exceeded\n");
    return draw_ctx->color_vect[0];
  }

  color.red = rgb_red;
  color.green = rgb_green;
  color.blue = rgb_blue;

  if (!gdk_colormap_alloc_color(draw_ctx->colormap, &color, TRUE, TRUE)) {
    printf("** Color not allocated !\n");
    color = glow_allocate_named_color(draw_ctx, "black");
  }

  draw_ctx->color_vect[draw_ctx->color_vect_cnt] = color;
  draw_ctx->color_vect_cnt++;

  return color;
}

static GdkColor glow_allocate_custom_color(GlowDrawGtk* draw_ctx,
    glow_eDrawType drawtype, int rgb_red, int rgb_green, int rgb_blue)
{
  GdkColor color;
  int idx = drawtype - glow_eDrawType_CustomColor1;

  if (!draw_ctx->get_customcolors())
    return color;

  color.red = rgb_red;
  color.green = rgb_green;
  color.blue = rgb_blue;

  if (!gdk_colormap_alloc_color(draw_ctx->colormap, &color, TRUE, TRUE)) {
    printf("** Color not allocated !\n");
    color = glow_allocate_named_color(draw_ctx, "black");
    return color;
  }

  if (draw_ctx->get_customcolors()->color_vect[idx].pixel != 0)
    gdk_colormap_free_colors(
        draw_ctx->colormap, &draw_ctx->get_customcolors()->color_vect[idx], 1);

  draw_ctx->get_customcolors()->color_vect[idx] = color;

  return color;
}

DrawWindGtk::DrawWindGtk()
    : DrawWind(), toplevel(0), window(0), background_pixmap(0)
{
  memset(clip_rectangle, 0, sizeof(clip_rectangle));
}

GlowDrawGtk::~GlowDrawGtk()
{
  closing_down = 1;

  ctx->set_nodraw();
  if (ctx->type() == glow_eCtxType_Grow)
    delete (GrowCtx*)ctx;
  else
    delete ctx;
  draw_free_gc(this);

  gdk_colormap_free_colors(colormap, color_vect, color_vect_cnt);

  if (m_wind.background_pixmap)
    g_object_unref(m_wind.background_pixmap);
  if (nav_wind.background_pixmap)
    g_object_unref(nav_wind.background_pixmap);

  if (timer_id)
    g_source_remove(timer_id);
}

void GlowDrawGtk::init_nav(GtkWidget* nav_widget)
{
  nav_wind.toplevel = nav_widget;
  nav_wind.window = nav_widget->window;

  gtk_widget_modify_bg(nav_widget, GTK_STATE_NORMAL, &background);

  ctx->no_nav = 0;
}

GlowDrawGtk::GlowDrawGtk(GtkWidget* toplevel, void** glow_ctx,
    int (*init_proc)(GtkWidget* w, GlowCtx* ctx, void* client_data),
    void* client_data, glow_eCtxType type)
    : ef(0), timer_id(0), click_sensitivity(0), color_vect_cnt(0),
      closing_down(0), customcolors_cnt(0)
{
  memset(gcs, 0, sizeof(gcs));
  memset(font, 0, sizeof(font));
  memset(cursors, 0, sizeof(cursors));
  memset(customcolors, 0, sizeof(customcolors));

  if (type == glow_eCtxType_Brow)
    ctx = (GlowCtx*)new BrowCtx("Claes context");
  else if (type == glow_eCtxType_Grow)
    ctx = (GlowCtx*)new GrowCtx("Claes context");
  else if (type == glow_eCtxType_ColPal)
    ctx = (GlowCtx*)new ColPalCtx("Claes context");
  else if (type == glow_eCtxType_Keyboard)
    ctx = (GlowCtx*)new KeyboardCtx("Claes context");
  else if (type == glow_eCtxType_Curve)
    ctx = (GlowCtx*)new CurveCtx("Claes context");
  else
    ctx = new GlowCtx("Claes context");
  *glow_ctx = (void*)ctx;

  ctx->gdraw = this;

  display = gtk_widget_get_display(toplevel);
  m_wind.toplevel = toplevel;
  m_wind.window = toplevel->window;
  screen = gtk_widget_get_screen(toplevel);

  ctx->mw.window = &m_wind;
  ctx->navw.window = &nav_wind;

  GtkStyle* style = gtk_widget_get_style(toplevel);
  background = style->bg[GTK_STATE_NORMAL];
  original_background = background;

  if (type == glow_eCtxType_Grow || type == glow_eCtxType_Curve
      || type == glow_eCtxType_Keyboard) {
    ctx->customcolors = create_customcolors();
    push_customcolors(ctx->customcolors);
  }

  colormap = gdk_colormap_new(gdk_visual_get_system(), TRUE);

  glow_create_gc(this, m_wind.window);

  glow_create_cursor(this);

  init_proc(toplevel, ctx, client_data);

  ctx->set_dirty();
}

void GlowDrawGtk::event_handler(GdkEvent event)
{
  static int button_pressed = 0;
  static int button_clicked = 0;
  static int button_clicked_and_pressed = 0;
  static int button1_pressed = 0;
  static int button2_pressed = 0;
  static int button3_pressed = 0;
  static int last_press_x = 0;
  static int last_press_y = 0;
  int sts = 1;

  if (closing_down)
    return;

  if (event.any.window == m_wind.window || event.type == GDK_KEY_PRESS) {
    if (GlowCtx::eventlog_enabled)
      log_event(&event);

    switch (event.type) {
    case GDK_KEY_PRESS: {
      guint keysym;

      keysym = event.key.keyval;
      if ((keysym >= 0x020 && keysym <= 0x20ac)
          || (keysym >= 0xFF80 && keysym <= 0xFFB9 && keysym != GDK_KP_Enter
                 && keysym != 0xFF44)) {
        unsigned char buff = event.key.string[0];

        if (buff >= 0x020)
          sts = ctx->event_handler(glow_eEvent_Key_Ascii, 0, 0, (int)buff, 0);
        else if (buff > 0) {
          sts = ctx->event_handler(
              glow_eEvent_Key_CtrlAscii, 0, 0, (int)buff, 0);
        }
      } else {
        switch (keysym) {
        case GDK_Return:
        case GDK_KP_Enter:
        case 0xFF44: // XK_KP_Enter sometimes...
          sts = ctx->event_handler(glow_eEvent_Key_Return, 0, 0, 0, 0);
          break;
        case GDK_Up:
          sts = ctx->event_handler(glow_eEvent_Key_Up, 0, 0, 0, 0);
          break;
        case GDK_Down:
          sts = ctx->event_handler(glow_eEvent_Key_Down, 0, 0, 0, 0);
          break;
        case GDK_Right:
          sts = ctx->event_handler(glow_eEvent_Key_Right, 0, 0, 0, 0);
          break;
        case GDK_Left:
          sts = ctx->event_handler(glow_eEvent_Key_Left, 0, 0, 0, 0);
          break;
        case GDK_Page_Up:
          sts = ctx->event_handler(glow_eEvent_Key_PageUp, 0, 0, 0, 0);
          break;
        case GDK_Page_Down:
          sts = ctx->event_handler(glow_eEvent_Key_PageDown, 0, 0, 0, 0);
          break;
        case GDK_Delete:
        case GDK_BackSpace:
          sts = ctx->event_handler(glow_eEvent_Key_BackSpace, 0, 0, 0, 0);
          break;
        case GDK_KP_F1:
          sts = ctx->event_handler(glow_eEvent_Key_PF1, 0, 0, 0, 0);
          break;
        case GDK_KP_F2:
          sts = ctx->event_handler(glow_eEvent_Key_PF2, 0, 0, 0, 0);
          break;
        case GDK_KP_F3:
          sts = ctx->event_handler(glow_eEvent_Key_PF3, 0, 0, 0, 0);
          break;
        case GDK_KP_F4:
          sts = ctx->event_handler(glow_eEvent_Key_PF4, 0, 0, 0, 0);
          break;
        case GDK_Cancel:
        case GDK_Escape:
          sts = ctx->event_handler(glow_eEvent_Key_Escape, 0, 0, 0, 0);
          break;
        case GDK_Tab:
          if (event.key.state & GDK_SHIFT_MASK)
            sts = ctx->event_handler(glow_eEvent_Key_ShiftTab, 0, 0, 0, 0);
          else
            sts = ctx->event_handler(glow_eEvent_Key_Tab, 0, 0, 0, 0);
          break;
        default:;
        }
      }
      break;
    }
    case GDK_BUTTON_PRESS:
      switch (event.button.button) {
      case 1:
        sts = ctx->event_handler(glow_eEvent_MB1Down, (int)event.button.x,
            (int)event.button.y, 0, 0);
        if (click_sensitivity & glow_mSensitivity_MB1Click
            && !(click_sensitivity & glow_mSensitivity_MB1DoubleClick)
            && !(click_sensitivity & glow_mSensitivity_MB1Press)) {
          memcpy(&last_event, &event, sizeof(event));
          button_pressed = 0;
          button_clicked = 1;
          last_press_x = (int)event.button.x;
          last_press_y = (int)event.button.y;
          return;
        } else if (!(click_sensitivity & glow_mSensitivity_MB1Click)
            && !(click_sensitivity & glow_mSensitivity_MB1DoubleClick)
            && click_sensitivity & glow_mSensitivity_MB1Press) {
          memcpy(&last_event, &event, sizeof(event));
          button_pressed = 1;
          button_clicked = 0;
          last_press_x = (int)event.button.x;
          last_press_y = (int)event.button.y;
        }
        break;
      case 3:
        sts = ctx->event_handler(glow_eEvent_MB3Down, (int)event.button.x,
            (int)event.button.y, 0, 0);
        if (click_sensitivity & glow_mSensitivity_MB3Press
            && !(click_sensitivity & glow_mSensitivity_MB3DoubleClick)
            && !(click_sensitivity & glow_mSensitivity_MB3Click)) {
          sts = ctx->event_handler(glow_eEvent_MB3Press, (int)event.button.x,
              (int)event.button.y, 0, 0);
          click_sensitivity = 0;
          return;
        } else if (click_sensitivity & glow_mSensitivity_MB3Click
            && !(click_sensitivity & glow_mSensitivity_MB3DoubleClick)
            && !(click_sensitivity & glow_mSensitivity_MB3Press)) {
          memcpy(&last_event, &event, sizeof(event));
          button_pressed = 0;
          button_clicked = 1;
          last_press_x = (int)event.button.x;
          last_press_y = (int)event.button.y;
          return;
        }
        break;
      }

      /* Detect press or click event */
      if (button_clicked) {
        /* Wait for release */
        button_clicked_and_pressed = 1;
        cancel_event_timer(this);
        button_clicked = 0;
        memcpy(&last_event, &event, sizeof(event));
        button_pressed = event.button.button;
        last_press_x = (int)event.button.x;
        last_press_y = (int)event.button.y;
        event_timer(this, 200);
        return;
      }
      if (!button_pressed) {
        memcpy(&last_event, &event, sizeof(event));
        button_pressed = event.button.button;
        last_press_x = (int)event.button.x;
        last_press_y = (int)event.button.y;
        event_timer(this, 200);
        return;
      } else {
        /* Press event, callback from timer */
        button_pressed = 0;
        button_clicked_and_pressed = 0;
        switch (event.button.button) {
        case 1:

          button1_pressed = 1;
          if ((event.button.state & GDK_SHIFT_MASK)
              && !(event.button.state & GDK_CONTROL_MASK)) {
            sts = ctx->event_handler(glow_eEvent_MB1PressShift,
                (int)event.button.x, (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          } else if (!(event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK)) {
            sts = ctx->event_handler(glow_eEvent_MB1PressCtrl,
                (int)event.button.x, (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          } else if ((event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK)) {
            sts = ctx->event_handler(glow_eEvent_MB1PressShiftCtrl,
                (int)event.button.x, (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          } else {
            sts = ctx->event_handler(glow_eEvent_MB1Press, (int)event.button.x,
                (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          }
          break;
        case 2:
          button2_pressed = 1;
          if ((event.button.state & GDK_SHIFT_MASK)
              && !(event.button.state & GDK_CONTROL_MASK)) {
            sts = ctx->event_handler(glow_eEvent_MB2PressShift,
                (int)event.button.x, (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          } else if (!(event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK)) {
            sts = ctx->event_handler(glow_eEvent_MB2PressShift,
                (int)event.button.x, (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          } else if ((event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK)) {
            sts = ctx->event_handler(glow_eEvent_MB2PressShiftCtrl,
                (int)event.button.x, (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          } else {
            sts = ctx->event_handler(glow_eEvent_MB2Press, (int)event.button.x,
                (int)event.button.y, 0, 0);
            click_sensitivity = 0;
          }
          break;
        case 3:
          button3_pressed = 1;
          sts = ctx->event_handler(glow_eEvent_MB3Press, (int)event.button.x,
              (int)event.button.y, 0, 0);
          click_sensitivity = 0;
          break;
        }
      }
      break;

    case GDK_BUTTON_RELEASE:
      button1_pressed = 0;
      button2_pressed = 0;
      button3_pressed = 0;

      switch (event.button.button) {
      case Button1:
        sts = ctx->event_handler(
            glow_eEvent_MB1Up, (int)event.button.x, (int)event.button.y, 0, 0);
        break;
      }

      if (!button_pressed) {
        if (button_clicked) {
          /* Button click, callback from timer */
          button_clicked = 0;
          switch (event.button.button) {
          case 1:
            if ((event.button.state & GDK_SHIFT_MASK)
                && !(event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB1ClickShift,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if (!(event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB1ClickCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if ((event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB1ClickShiftCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else {
              sts = ctx->event_handler(glow_eEvent_MB1Click,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            }
            break;
          case 2:
            if ((event.button.state & GDK_SHIFT_MASK)
                && !(event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB2ClickShift,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if (!(event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB2ClickCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if ((event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB2ClickShiftCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else {
              sts = ctx->event_handler(glow_eEvent_MB2Click,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            }
            break;
          case 3:
            sts = ctx->event_handler(glow_eEvent_MB3Click, (int)event.button.x,
                (int)event.button.y, 0, 0);
            if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
              return;
            click_sensitivity = 0;
            break;
          }
        } else {
          /* Button release */
          sts = ctx->event_handler(glow_eEvent_ButtonRelease,
              (int)event.button.x, (int)event.button.y, 0, 0);
          if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
            return;
        }
      } else {
        /* Button click */
        cancel_event_timer(this);
        if (!button_clicked_and_pressed) {
          /* wait for button double click */
          memcpy(&last_event, &event, sizeof(event));
          button_clicked = 1;
          event_timer(this, 200);
          button_pressed = 0;
          return;
        } else {
          /* Button double click */
          cancel_event_timer(this);
          button_clicked = 0;
          button_pressed = 0;
          button_clicked_and_pressed = 0;
          switch (event.button.button) {
          case 1:
            if ((event.button.state & GDK_SHIFT_MASK)
                && !(event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB1DoubleClickShift,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if (!(event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB1DoubleClickCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if ((event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB1DoubleClickShiftCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else {
              sts = ctx->event_handler(glow_eEvent_MB1DoubleClick,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            }
            break;
          case 2:
            if ((event.button.state & GDK_SHIFT_MASK)
                && !(event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB2DoubleClickShift,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if (!(event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB2DoubleClickCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else if ((event.button.state & GDK_SHIFT_MASK)
                && (event.button.state & GDK_CONTROL_MASK)) {
              sts = ctx->event_handler(glow_eEvent_MB2DoubleClickShiftCtrl,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            } else {
              sts = ctx->event_handler(glow_eEvent_MB2DoubleClick,
                  (int)event.button.x, (int)event.button.y, 0, 0);
              if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
                return;
              click_sensitivity = 0;
            }
            break;
          case 3:
            break;
          }
        }
      }
      break;
    case GDK_EXPOSE:
      ctx->is_dirty = 1;
      sts = ctx->event_handler(glow_eEvent_Exposure, event.expose.area.x,
          event.expose.area.y, event.expose.area.width,
          event.expose.area.height);
      begin(&ctx->mw);
      ctx->draw(&ctx->mw, event.expose.area.x, event.expose.area.y,
          event.expose.area.x + event.expose.area.width,
          event.expose.area.y + event.expose.area.height);
      end();
      ctx->is_dirty = 0;
      break;
    case GDK_VISIBILITY_NOTIFY:
      switch (event.visibility.state) {
      case GDK_VISIBILITY_UNOBSCURED:
        sts = ctx->event_handler(glow_eEvent_VisibilityUnobscured, 0, 0, 0, 0);
        break;
      default:
        sts = ctx->event_handler(glow_eEvent_VisibilityObscured, 0, 0, 0, 0);
        break;
      }
      break;
    case GDK_MOTION_NOTIFY:
      if (event.motion.is_hint) {
        int x, y;

        gdk_window_get_pointer(event.any.window, &x, &y, NULL);
        event.button.x = x;
        event.button.y = y;
      }

      if (button3_pressed) {
        button3_pressed = 0;
        button_pressed = 0;
        button_clicked_and_pressed = 0;
      }
      if (button_pressed
          && (ABS(event.button.x - last_press_x) > DRAW_PRESS_PIX
                 || ABS(event.button.y - last_press_y) > DRAW_PRESS_PIX)) {

        event.button.x = last_press_x;
        event.button.y = last_press_y;

        /* Button press */
        cancel_event_timer(this);
        switch (button_pressed) {
        case 1:
          button1_pressed = 1;
          if ((event.button.state & GDK_SHIFT_MASK)
              && !(event.button.state & GDK_CONTROL_MASK))
            sts = ctx->event_handler(glow_eEvent_MB1PressShift,
                (int)event.button.x, (int)event.button.y, 0, 0);
          else if (!(event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK))
            sts = ctx->event_handler(glow_eEvent_MB1PressCtrl,
                (int)event.button.x, (int)event.button.y, 0, 0);
          else if ((event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK))
            sts = ctx->event_handler(glow_eEvent_MB1PressShiftCtrl,
                (int)event.button.x, (int)event.button.y, 0, 0);
          else
            sts = ctx->event_handler(glow_eEvent_MB1Press, (int)event.button.x,
                (int)event.button.y, 0, 0);
          break;
        case 2:
          button2_pressed = 1;
          if ((event.button.state & GDK_SHIFT_MASK)
              && !(event.button.state & GDK_CONTROL_MASK))
            sts = ctx->event_handler(glow_eEvent_MB2PressShift,
                (int)event.button.x, (int)event.button.y, 0, 0);
          else if (!(event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK))
            sts = ctx->event_handler(glow_eEvent_MB2PressShift,
                (int)event.button.x, (int)event.button.y, 0, 0);
          else if ((event.button.state & GDK_SHIFT_MASK)
              && (event.button.state & GDK_CONTROL_MASK))
            sts = ctx->event_handler(glow_eEvent_MB2PressShiftCtrl,
                (int)event.button.x, (int)event.button.y, 0, 0);
          else
            sts = ctx->event_handler(glow_eEvent_MB2Press, (int)event.button.x,
                (int)event.button.y, 0, 0);
          break;
        case 3:
          button3_pressed = 1;
          sts = ctx->event_handler(glow_eEvent_MB3Press, (int)event.button.x,
              (int)event.button.y, 0, 0);
          break;
        }
        button_pressed = 0;
        button_clicked_and_pressed = 0;
      }
      if (button1_pressed || button2_pressed || button3_pressed)
        sts = ctx->event_handler(glow_eEvent_ButtonMotion, (int)event.button.x,
            (int)event.button.y, 0, 0);
      else
        sts = ctx->event_handler(glow_eEvent_CursorMotion, (int)event.button.x,
            (int)event.button.y, 0, 0);
      break;
    case GDK_ENTER_NOTIFY:
      sts = ctx->event_handler(glow_eEvent_Enter, (int)event.crossing.x,
          (int)event.crossing.y, 0, 0);
      break;
    case GDK_LEAVE_NOTIFY:

      sts = ctx->event_handler(glow_eEvent_Leave, (int)event.crossing.x,
          (int)event.crossing.y, 0, 0);
      break;
    case GDK_MAP:
      sts = ctx->event_handler(glow_eEvent_Map, 0, 0, 0, 0);
      break;
    case GDK_UNMAP:
      sts = ctx->event_handler(glow_eEvent_Unmap, 0, 0, 0, 0);
      break;
    case GDK_FOCUS_CHANGE:
      break;
    case GDK_SCROLL:
      if (event.scroll.direction == GDK_SCROLL_UP)
        sts = ctx->event_handler(glow_eEvent_ScrollUp, (int)event.scroll.x,
            (int)event.scroll.y, 0, 0);
      else if (event.scroll.direction == GDK_SCROLL_DOWN)
        sts = ctx->event_handler(glow_eEvent_ScrollDown, (int)event.scroll.x,
            (int)event.scroll.y, 0, 0);
      break;
    default:
      break;
    }
  } else if (event.any.window == nav_wind.window) {
    switch (event.type) {
    case GDK_BUTTON_PRESS:
      switch (event.button.button) {
      case 1:
        button1_pressed = 1;
        ctx->event_handler_nav(
            glow_eEvent_MB1Press, (int)event.button.x, (int)event.button.y);
        break;
      case 2:
        button2_pressed = 1;
        ctx->event_handler_nav(
            glow_eEvent_MB2Press, (int)event.button.x, (int)event.button.y);
        break;
      case 3:
        button3_pressed = 1;
        break;
      }
      break;
    case GDK_BUTTON_RELEASE:
      switch (event.button.button) {
      case 1:
        button1_pressed = 0;
        break;
      case 2:
        button2_pressed = 0;
        break;
      case 3:
        button3_pressed = 0;
        break;
      }
      sts = ctx->event_handler_nav(
          glow_eEvent_ButtonRelease, (int)event.button.x, (int)event.button.y);
      break;
    case GDK_EXPOSE:
      ctx->is_dirty = 1;
      sts = ctx->event_handler_nav(glow_eEvent_Exposure, 0, 0);
      begin(&ctx->navw);
      ctx->draw(&ctx->navw, event.expose.area.x, event.expose.area.y,
          event.expose.area.x + event.expose.area.width,
          event.expose.area.y + event.expose.area.height);
      end();
      ctx->is_dirty = 0;
      break;
    case GDK_MOTION_NOTIFY:
      if (event.motion.is_hint) {
        int x, y;

        gdk_window_get_pointer(event.any.window, &x, &y, NULL);
        event.button.x = x;
        event.button.y = y;
      }
      if (button1_pressed || button2_pressed || button3_pressed)
        sts = ctx->event_handler_nav(
            glow_eEvent_ButtonMotion, (int)event.button.x, (int)event.button.y);
      else
        sts = ctx->event_handler_nav(
            glow_eEvent_CursorMotion, (int)event.button.x, (int)event.button.y);
      break;
    default:;
    }
  }
  if (sts == GLOW__TERMINATED || sts == GLOW__DESTROYED)
    return;

  gdk_display_flush(display);
  return;
}

void GlowDrawGtk::enable_event(glow_eEvent event, glow_eEventType event_type,
    int (*event_cb)(GlowCtx* ctx, glow_tEvent event))
{
  ctx->enable_event(event, event_type, event_cb);
}

int GlowDrawGtk::begin(GlowWind *w) {
  if (!this->w) {
    this->w = w;
    return 1;
  }
  return 0;
}

void GlowDrawGtk::end() {
  this->w = NULL;
}

void GlowDrawGtk::set_dirty(DrawWind* wind) {
  GdkWindow *w = ((DrawWindGtk*)wind)->window;
  GdkRectangle r = {0, 0, gdk_window_get_width(w), gdk_window_get_height(w)};
  gdk_window_invalidate_rect(w, &r, false);
}

void GlowDrawGtk::rect(int x, int y, int width, int height,
    glow_eDrawType gc_type, int fill, int idx, int highlight)
{
  if (gc_type == glow_eDrawType_LineGray && highlight)
    gc_type = glow_eDrawType_Line;

  set_clip(get_gc(this, gc_type + highlight, idx));

  gdk_draw_rectangle(((DrawWindGtk*)w->window)->window,
      get_gc(this, gc_type + highlight, idx), fill, x, y, width, height);

  reset_clip(get_gc(this, gc_type + highlight, idx));
}

void GlowDrawGtk::arrow(int x1, int y1, int x2, int y2, int x3,
    int y3, glow_eDrawType gc_type, int idx, int highlight)
{
  GdkPoint p[4] = { { x1, y1 }, { x2, y2 }, { x3, y3 }, { x1, y1 } };

  if (gc_type == glow_eDrawType_LineGray && highlight)
    gc_type = glow_eDrawType_Line;

  set_clip(get_gc(this, gc_type + highlight, idx));

  gdk_draw_polygon(((DrawWindGtk*)w->window)->window,
      get_gc(this, gc_type + highlight, idx), 1, p, 4);

  reset_clip(get_gc(this, gc_type + highlight, idx));
}

void GlowDrawGtk::arc(int x, int y, int width, int height, int angle1,
    int angle2, glow_eDrawType gc_type, int fill, int idx, int highlight)
{
  if (gc_type == glow_eDrawType_LineGray && highlight)
    gc_type = glow_eDrawType_Line;

  // Fix for highlight for connections in grow
  if (highlight && ctx->type() == glow_eCtxType_Grow)
    gc_type = glow_eDrawType_LineHighlight;

  if (angle1 >= 360)
    angle1 = angle1 - angle1 / 360 * 360;
  else if (angle1 < 0)
    angle1 = angle1 + (-angle1 / 360 + 1) * 360;

  set_clip(get_gc(this, gc_type + highlight, idx));

  gdk_draw_arc(((DrawWindGtk*)w->window)->window,
      get_gc(this, gc_type + highlight, idx), fill, x, y, width, height,
      angle1 * 64, angle2 * 64);

  reset_clip(get_gc(this, gc_type + highlight, idx));
}

void GlowDrawGtk::line(int x1, int y1, int x2, int y2, glow_eDrawType gc_type,
    int idx, int highlight, glow_eLineType line_type)
{
  // Fix for highlight for connections in grow
  if (highlight && ctx->type() == glow_eCtxType_Grow)
    gc_type = glow_eDrawType_LineHighlight;

  if (gc_type == glow_eDrawType_LineGray && highlight)
    gc_type = glow_eDrawType_Line;

  if (line_type != glow_eLineType_Solid) {
    gdk_gc_set_line_attributes(get_gc(this, gc_type + highlight, idx), idx + 1,
        GDK_LINE_ON_OFF_DASH, GDK_CAP_BUTT, GDK_JOIN_MITER);
    gint dash_offset = 0;
    gint8 dashes[4];
    switch (line_type) {
    case glow_eLineType_Dashed1:
      dashes[0] = 1 + idx;
      gdk_gc_set_dashes(
          get_gc(this, gc_type + highlight, idx), dash_offset, dashes, 1);
      break;
    case glow_eLineType_Dashed2:
      dashes[0] = 1 + 2 * idx;
      gdk_gc_set_dashes(
          get_gc(this, gc_type + highlight, idx), dash_offset, dashes, 1);
      break;
    case glow_eLineType_Dashed3:
      dashes[0] = 1 + 3 * idx;
      gdk_gc_set_dashes(
          get_gc(this, gc_type + highlight, idx), dash_offset, dashes, 1);
      break;
    case glow_eLineType_Dotted: {
      dashes[0] = 1 + idx;
      dashes[1] = 1 + 4 * idx;
      gdk_gc_set_dashes(
          get_gc(this, gc_type + highlight, idx), dash_offset, dashes, 2);
      break;
    }
    case glow_eLineType_DotDashed1: {
      dashes[0] = 1 + 3 * idx;
      dashes[1] = 1 + 2 * idx;
      dashes[2] = 1 + idx;
      dashes[3] = 1 + 2 * idx;
      gdk_gc_set_dashes(
          get_gc(this, gc_type + highlight, idx), dash_offset, dashes, 4);
      break;
    }
    case glow_eLineType_DotDashed2: {
      dashes[0] = 1 + 6 * idx;
      dashes[1] = 1 + 3 * idx;
      dashes[2] = 1 + idx;
      dashes[3] = 1 + 3 * idx;
      gdk_gc_set_dashes(
          get_gc(this, gc_type + highlight, idx), dash_offset, dashes, 4);
      break;
    }
    default:;
    }
  }

  set_clip(get_gc(this, gc_type + highlight, idx));

  gdk_draw_line(((DrawWindGtk*)w->window)->window,
      get_gc(this, gc_type + highlight, idx), x1, y1, x2, y2);

  reset_clip(get_gc(this, gc_type + highlight, idx));

  gdk_gc_set_line_attributes(get_gc(this, gc_type + highlight, idx), idx + 1,
      GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
}

void GlowDrawGtk::polyline(glow_sPointX* points, int point_cnt,
    glow_eDrawType gc_type, int fill, int idx, int highlight)
{
  if (gc_type == glow_eDrawType_LineGray && highlight)
    gc_type = glow_eDrawType_Line;

  set_clip(get_gc(this, gc_type + highlight, idx));

  GdkPoint* gpoints;
  if (fill) {
    gpoints = points_to_gdk_points(points, point_cnt);
    gdk_draw_polygon(((DrawWindGtk*)w->window)->window,
        get_gc(this, gc_type + highlight, 0), fill, gpoints, point_cnt);
  } else {
    int cnt;
    gpoints = points_to_gdk_points_curve(points, point_cnt, &cnt);
    gdk_draw_lines(((DrawWindGtk*)w->window)->window,
        get_gc(this, gc_type + highlight, idx), gpoints, cnt);
  }
  free(gpoints);

  reset_clip(get_gc(this, gc_type + highlight, idx));
}

#define FONT_SCALE 0.7
#define FONT_DESCENT 0.22

static char* font_string(int font_idx, int font_type, double size)
{
  static char str[80];
  char bold_str[20];

  if (font_type == glow_eDrawFont_HelveticaBold)
    strcpy(bold_str, "Bold ");
  else
    strcpy(bold_str, "");

  switch (font_idx) {
  case glow_eFont_Helvetica:
    sprintf(str, "%s %s%f", "Helvetica", bold_str, FONT_SCALE * size);
    break;
  case glow_eFont_Times:
    sprintf(str, "%s %s%f", "DejaVu Serif", bold_str, FONT_SCALE * size);
    break;
  case glow_eFont_NewCenturySchoolbook:
    sprintf(
        str, "%s %s%f", "Century Schoolbook L", bold_str, FONT_SCALE * size);
    break;
  case glow_eFont_Courier:
    sprintf(str, "%s %s%f", "Courier 10 pitch", bold_str, FONT_SCALE * size);
    break;
  default:
    sprintf(str, "%s %s%f", "Lucida Sans", bold_str, FONT_SCALE * size);
  }
  for (char* s = str; *s; s++)
    if (*s == ',')
      *s = '.';
  return str;
}

void GlowDrawGtk::text(int x, int y, char* text, int len,
    glow_eDrawType gc_type, glow_eDrawType color, int idx, int highlight,
    glow_eFont font_idx, double size, int rot)
{
  if (font_idx >= glow_eFont__)
    font_idx = glow_eFont_Helvetica;

  int font_type = get_font_type(gc_type);

  set_clip(get_gc(this, gc_type, idx));

  if (color != glow_eDrawType_Line) {
    GdkGCValues xgcv;
    gdk_gc_get_values(get_gc(this, color, 0), &xgcv);
    gdk_gc_set_values(get_gc(this, gc_type, idx), &xgcv, GDK_GC_FOREGROUND);
  }

  PangoRenderer* pr = gdk_pango_renderer_get_default(screen);
  gdk_pango_renderer_set_gc(GDK_PANGO_RENDERER(pr), get_gc(this, gc_type, idx));

  gdk_pango_renderer_set_drawable(GDK_PANGO_RENDERER(pr), ((DrawWindGtk*)w->window)->window);

  PangoContext* pctx = gdk_pango_context_get_for_screen(screen);
  PangoLayout* layout = pango_layout_new(pctx);
  if (((GrowCtx*)ctx)->text_coding != glow_eTextCoding_UTF_8) {
    char* textutf8
        = g_convert(text, -1, "UTF-8", "ISO8859-1", NULL, NULL, NULL);
    pango_layout_set_text(layout, textutf8, -1);
    g_free(textutf8);
  } else
    pango_layout_set_text(layout, text, -1);

  if (rot != 0) {
    rot = 360 - rot;
    PangoContext* pc = pango_layout_get_context(layout);
    pango_context_set_base_gravity(pc, PANGO_GRAVITY_SOUTH);
    PangoMatrix lpm = PANGO_MATRIX_INIT;
    pango_matrix_rotate(&lpm, (double)rot);
    pango_context_set_matrix(pc, &lpm);
    pango_layout_context_changed(layout);
  }

  PangoFontDescription* desc = pango_font_description_from_string(
      font_string(font_idx, font_type, size));
  pango_layout_set_font_description(layout, desc);

  int width, height;
  pango_layout_get_size(layout, &width, &height);
  height = (int)(height * 0.9);

  int px, py;
  if (rot == 180) {
    px = -PANGO_SCALE * x - width;
    py = (int)(-PANGO_SCALE * y - FONT_DESCENT * height);
  } else if (rot == 90) {
    px = -PANGO_SCALE * y - width + height / 2;
    py = PANGO_SCALE * x;
  } else if (rot == 270) {
    px = PANGO_SCALE * y - width;
    py = -PANGO_SCALE * x - height;
  } else {
    px = PANGO_SCALE * x;
    py = (int)(PANGO_SCALE * y - (1.0 - FONT_DESCENT) * height);
  }
  pango_renderer_draw_layout(pr, layout, px, py);

  if (rot != 0) {
    PangoContext* pc = pango_layout_get_context(layout);
    pango_context_set_matrix(pc, 0);
  }
  g_object_unref(layout);
  pango_font_description_free(desc);

  gdk_pango_renderer_set_drawable(GDK_PANGO_RENDERER(pr), 0);
  gdk_pango_renderer_set_gc(GDK_PANGO_RENDERER(pr), 0);

  reset_clip(get_gc(this, gc_type, idx));

  if (color != glow_eDrawType_Line) {
    GdkGCValues xgcv;
    gdk_gc_get_values(get_gc(this, glow_eDrawType_Line, 0), &xgcv);
    gdk_gc_set_values(get_gc(this, gc_type, idx), &xgcv, GDK_GC_FOREGROUND);
  }
}

void GlowDrawGtk::text_cursor(int x, int y, char* text, int len,
    glow_eDrawType gc_type, glow_eDrawType color, int idx, int highlight,
    int pos, glow_eFont font, double size)
{
  int theight, tdescent, width, height, descent;
  get_text_extent(
      "A", 1, gc_type, idx, font, &width, &height, &descent, size, 0);
  if (pos != 0)
    get_text_extent(
        text, pos, gc_type, idx, font, &width, &theight, &tdescent, size, 0);
  else
    width = 0;

  set_clip(get_gc(this, gc_type, idx));

  gdk_draw_line(((DrawWindGtk*)w->window)->window,
      get_gc(this, color, 1), x + width, y + descent, x + width,
      y - height + descent);

  reset_clip(get_gc(this, gc_type, idx));
}

void GlowDrawGtk::set_image_clip_mask(glow_tPixmap pixmap, int x, int y)
{
  gdk_gc_set_clip_mask(
      get_gc(this, glow_eDrawType_Line, 0), (GdkPixmap*)pixmap);
  gdk_gc_set_clip_origin(get_gc(this, glow_eDrawType_Line, 0), x, y);
}

void GlowDrawGtk::reset_image_clip_mask()
{
  gdk_gc_set_clip_mask(get_gc(this, glow_eDrawType_Line, 0), NULL);
  gdk_gc_set_clip_origin(get_gc(this, glow_eDrawType_Line, 0), 0, 0);
}

void GlowDrawGtk::image(int x, int y, int width,
    int height, glow_tImImage image, glow_tPixmap pixmap,
    glow_tPixmap clip_mask)
{
  cairo_t* cr = gdk_cairo_create(((DrawWindGtk*)w->window)->window);

  if (clip_mask)
    set_image_clip_mask(clip_mask, x, y);
  else {
    set_clip(get_gc(this, glow_eDrawType_Line, 0));
    set_cairo_clip(cr);
  }

  cairo_t* cr2 = gdk_cairo_create(((DrawWindGtk*)w->window)->window);
  gdk_cairo_set_source_pixbuf(cr2, (GdkPixbuf*)image, x, y);
  cairo_paint(cr2);
  cairo_destroy(cr2);

  if (clip_mask)
    reset_image_clip_mask();
  else {
    reset_clip(get_gc(this, glow_eDrawType_Line, 0));
    reset_cairo_clip( cr);
  }
  cairo_destroy(cr);
}

void GlowDrawGtk::clear()
{
  DrawWindGtk* w = ((DrawWindGtk*)this->w->window);
  if (w->background_pixmap) {
    set_clip(get_gc(this, glow_eDrawType_Line, 0));

    if (!((GrowCtx*)ctx)->background_tiled)
      gdk_draw_drawable(w->window,
          get_gc(this, glow_eDrawType_Line, 0), w->background_pixmap, 0, 0, 0, 0,
          this->w->window_width, this->w->window_height);
    else {
      int background_width, background_height;
      gdk_drawable_get_size(w->background_pixmap, &background_width,
          &background_height);
      for (int i = 0; i <= this->w->window_width / background_width; i++)
        for (int j = 0; j <= this->w->window_height / background_height; j++)
          gdk_draw_drawable(w->window,
              get_gc(this, glow_eDrawType_Line, 0), w->background_pixmap, 0, 0,
              i * background_width, j * background_height,
              background_width, background_height);
    }
    reset_clip(get_gc(this, glow_eDrawType_Line, 0));
  } else {
    glow_eDrawType bg = ((GrowCtx*)ctx)->background_color;
    set_clip(get_gc(this, bg, 0));

    gdk_draw_rectangle(w->window, get_gc(this, bg, 0), 1, 0, 0,
        this->w->window_width, this->w->window_height);
    reset_clip(get_gc(this, bg, 0));
  }
}

void GlowDrawGtk::get_window_size(DrawWind* wind, int* width, int* height)
{
  DrawWindGtk* w = (DrawWindGtk*)wind;
  gdk_drawable_get_size(w->window, width, height);
}

void GlowDrawGtk::set_window_size(DrawWind* wind, int width, int height)
{
  DrawWindGtk* w = (DrawWindGtk*)wind;
  if (!w->window)
    return;

  gdk_window_resize(w->window, width, height);
}

static gboolean draw_timer_cb(void* data)
{
  draw_sTimerCb* timer_cb = (draw_sTimerCb*)data;
  GlowDrawGtk* draw_ctx = (GlowDrawGtk*)timer_cb->ctx->gdraw;

  (timer_cb->callback_func)(timer_cb->ctx);

  gdk_display_flush(draw_ctx->display);
  free(timer_cb);
  return FALSE;
}

static gboolean event_timer_cb(void* ctx)
{
  GlowDrawGtk* draw_ctx = (GlowDrawGtk*)ctx;
  draw_ctx->timer_id = 0;

  draw_ctx->event_handler(last_event);
  return FALSE;
}

static void cancel_event_timer(GlowDrawGtk* draw_ctx)
{
  if (draw_ctx->timer_id) {
    g_source_remove(draw_ctx->timer_id);
    draw_ctx->timer_id = 0;
  }
}

static void event_timer(GlowDrawGtk* draw_ctx, int time_ms)
{
  draw_ctx->timer_id = g_timeout_add(time_ms, event_timer_cb, draw_ctx);
}

void GlowDrawGtk::set_timer(
    GlowCtx* gctx, int time_ms, void (*callback_func)(GlowCtx* ctx), void** id)
{
  draw_sTimerCb* timer_cb;

  timer_cb = (draw_sTimerCb*)calloc(1, sizeof(draw_sTimerCb));
  timer_cb->ctx = gctx;
  timer_cb->callback_func = callback_func;

  timer_cb->timer_id = g_timeout_add(time_ms, draw_timer_cb, timer_cb);

  *id = (void*)timer_cb;
}

void GlowDrawGtk::remove_timer(void* id)
{
  g_source_remove(((draw_sTimerCb*)id)->timer_id);

  free((char*)id);
}

void GlowDrawGtk::set_cursor(DrawWind* wind, glow_eDrawCursor cursor)
{
  DrawWindGtk* w = (DrawWindGtk*)wind;
  if (!w->window)
    return;

  if (cursor == glow_eDrawCursor_Normal)
    gdk_window_set_cursor(w->window, NULL);
  else
    gdk_window_set_cursor(w->window, cursors[cursor]);
  gdk_display_flush(display);
}

void GlowDrawGtk::get_text_extent(const char* text, int len,
    glow_eDrawType gc_type, int idx, glow_eFont font_idx, int* width,
    int* height, int* descent, double size, int rot)
{
  DrawWindGtk* w = &m_wind;

  if (font_idx >= glow_eFont__)
    font_idx = glow_eFont_Helvetica;

  if (gc_type == glow_eDrawType_TextHelvetica)
    gc_type = glow_eDrawType_TextHelveticaErase;
  else if (gc_type == glow_eDrawType_TextHelveticaBold)
    gc_type = glow_eDrawType_TextHelveticaEraseBold;

  int font_type = get_font_type(gc_type);

  PangoRenderer* pr = gdk_pango_renderer_get_default(screen);
  gdk_pango_renderer_set_gc(GDK_PANGO_RENDERER(pr), gcs[gc_type][0]);
  gdk_pango_renderer_set_drawable(GDK_PANGO_RENDERER(pr), w->window);

  PangoContext* pctx = gdk_pango_context_get_for_screen(screen);
  PangoLayout* layout = pango_layout_new(pctx);
  if (((GrowCtx*)ctx)->text_coding != glow_eTextCoding_UTF_8) {
    char* textutf8
        = g_convert(text, len, "UTF-8", "ISO8859-1", NULL, NULL, NULL);
    pango_layout_set_text(layout, textutf8, -1);
    g_free(textutf8);
  } else {
    char* textp = (char*)calloc(1, len + 1);
    strncpy(textp, text, len);
    pango_layout_set_text(layout, text, -1);
    free(textp);
  }

  PangoFontDescription* desc = pango_font_description_from_string(
      font_string(font_idx, font_type, size));
  pango_layout_set_font_description(layout, desc);

  int lwidth, lheight;
  pango_layout_get_size(layout, &lwidth, &lheight);
  lheight = (int)(lheight * 0.9);

  if (rot == 90 || rot == 270) {
    *height = lwidth / PANGO_SCALE;
    *width = lheight / PANGO_SCALE;
  } else {
    *width = lwidth / PANGO_SCALE;
    *height = lheight / PANGO_SCALE;
  }
  *descent = (int)(FONT_DESCENT * lheight / PANGO_SCALE);

  g_object_unref(layout);
  pango_font_description_free(desc);
  gdk_pango_renderer_set_drawable(GDK_PANGO_RENDERER(pr), 0);
  gdk_pango_renderer_set_gc(GDK_PANGO_RENDERER(pr), 0);
}

static int glow_read_color_file(
    const char* filename, draw_sColor** color_array, int* size)
{
  char line[80];
  draw_sColor* color_p;
  ifstream fp;
  int nr;
  int line_cnt;
  float f_red, f_green, f_blue;

  if (!check_file(filename)) {
    return 0;
  } else {
    printf("** Opening color file %s\n", filename);

    fp.open(filename);
    if (!fp)
      return GLOW__FILEOPEN;

    *color_array = (draw_sColor*)calloc(300, sizeof(draw_sColor));
    *size = 0;
    line_cnt = 0;
    color_p = *color_array;
    while (*size < 300) {
      fp.getline(line, sizeof(line));
      if (line[0] == 0)
        break;
      line_cnt++;
      if (line[0] == '!' || line[0] == '#')
        continue;
      nr = sscanf(line, "%f %f %f", &f_red, &f_green, &f_blue);
      if (nr != 3)
        printf("** Syntax error in file %s, line %d", filename, line_cnt);
      else {
        color_p->red = int(f_red * 65535);
        color_p->green = int(f_green * 65535);
        color_p->blue = int(f_blue * 65535);
        color_p++;
        (*size)++;
      }
    }
    fp.close();
  }

  return 1;
}

void GlowDrawGtk::set_background(DrawWind* wind, glow_eDrawType drawtype,
    glow_tPixmap pixmap, glow_tImImage image, int pixmap_width,
    int pixmap_height)
{
  DrawWindGtk* w = (DrawWindGtk*)wind;
  GdkGCValues xgcv;
  int i;
  double r, g, b;

  if (drawtype == glow_eDrawType_LineErase)
    drawtype = glow_eDrawType_Color32;

  GlowColor::rgb_color((int)drawtype, &r, &g, &b, get_customcolors());
  background = glow_allocate_color(
      this, int(r * 65535), int(g * 65535), int(b * 65535));

  // Change erase gcs
  xgcv.foreground = background;
  xgcv.background = background;
  for (i = 0; i < DRAW_TYPE_SIZE; i++) {
    gdk_gc_set_values(get_gc(this, glow_eDrawType_LineErase, i), &xgcv,
        (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
    gdk_gc_set_values(get_gc(this, glow_eDrawType_TextHelveticaErase, i), &xgcv,
        (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
    gdk_gc_set_values(get_gc(this, glow_eDrawType_TextHelveticaEraseBold, i),
        &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
  }

  if (!image) {
    gtk_widget_modify_bg(w->toplevel, GTK_STATE_NORMAL, &background);
  } else {
    GdkBitmap* mask;

    if (w->background_pixmap)
      g_object_unref(w->background_pixmap);

    gdk_pixbuf_render_pixmap_and_mask_for_colormap(
        (GdkPixbuf*)image, colormap, &w->background_pixmap, &mask, 0);
    gdk_window_set_back_pixmap(w->window, w->background_pixmap, FALSE);
  }
}

void GlowDrawGtk::reset_background(DrawWind* wind)
{
  GdkGCValues xgcv;

  background = original_background;
  gtk_widget_modify_bg(m_wind.toplevel, GTK_STATE_NORMAL, &background);

  // Change erase gcs
  xgcv.foreground = background;
  xgcv.background = background;
  for (int i = 0; i < DRAW_TYPE_SIZE; i++) {
    gdk_gc_set_values(get_gc(this, glow_eDrawType_LineErase, i), &xgcv,
        (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
    gdk_gc_set_values(get_gc(this, glow_eDrawType_TextHelveticaErase, i), &xgcv,
        (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
    gdk_gc_set_values(get_gc(this, glow_eDrawType_TextHelveticaEraseBold, i),
        &xgcv, (GdkGCValuesMask)(GDK_GC_FOREGROUND | GDK_GC_BACKGROUND));
  }
}

void GlowDrawGtk::set_clip(GdkGC* gc)
{
  DrawWindGtk* w = (DrawWindGtk*)this->w->window;
  if (w->clip_cnt > 0) {
    gdk_gc_set_clip_rectangle(gc, &w->clip_rectangle[w->clip_cnt - 1]);
  }
}

void GlowDrawGtk::reset_clip(GdkGC* gc)
{
  DrawWindGtk* w = (DrawWindGtk*)this->w->window;
  if (w->clip_cnt > 0) {
    gdk_gc_set_clip_rectangle(gc, NULL);
  }
}

int GlowDrawGtk::set_clip_rectangle(
    DrawWind* wind, int ll_x, int ll_y, int ur_x, int ur_y)
{
  DrawWindGtk* w = (DrawWindGtk*)wind;
  if (w->clip_cnt >= DRAW_CLIP_SIZE) {
    printf("** Draw clip refused\n");
    return 0;
  }
  int x0 = MIN(ll_x, ur_x);
  int x1 = MAX(ll_x, ur_x);
  int y0 = MIN(ll_y, ur_y);
  int y1 = MAX(ll_y, ur_y);
  if (w->clip_cnt > 0) {
    x0 = MAX(x0, w->clip_rectangle[w->clip_cnt - 1].x);
    x1 = MIN(x1, w->clip_rectangle[w->clip_cnt - 1].x
            + w->clip_rectangle[w->clip_cnt - 1].width);
    y0 = MAX(y0, w->clip_rectangle[w->clip_cnt - 1].y);
    y1 = MIN(y1, w->clip_rectangle[w->clip_cnt - 1].y
            + w->clip_rectangle[w->clip_cnt - 1].height);
  }
  w->clip_rectangle[w->clip_cnt].x = x0;
  w->clip_rectangle[w->clip_cnt].y = y0;
  w->clip_rectangle[w->clip_cnt].width = x1 - x0;
  w->clip_rectangle[w->clip_cnt].height = y1 - y0;
  w->clip_cnt++;
  return 1;
}

void GlowDrawGtk::draw_point(
    int x1, int y1, glow_eDrawType gc_type)
{
  set_clip(get_gc(this, gc_type, 0));

  gdk_draw_point(((DrawWindGtk*)w->window)->window, get_gc(this, gc_type, 0), x1, y1);

  reset_clip(get_gc(this, gc_type, 0));
}

void GlowDrawGtk::draw_points(glow_sPointX* points,
    int point_num, glow_eDrawType gc_type, int idx)
{
  set_clip(get_gc(this, gc_type, 0));

  if (idx > 0) {
    for (int i = 0; i < point_num; i++) {
      gdk_draw_rectangle(((DrawWindGtk*)w->window)->window,
          get_gc(this, gc_type, 0), 1, points[i].x - idx / 2,
          points[i].y - idx / 2, idx, idx);
    }
    reset_clip(get_gc(this, gc_type, 0));
    return;
  }

  GdkPoint* gpoints = points_to_gdk_points(points, point_num);
  gdk_draw_points(((DrawWindGtk*)w->window)->window, get_gc(this, gc_type, 0),
      gpoints, point_num);
  free(gpoints);

  reset_clip(get_gc(this, gc_type, 0));
}

void GlowDrawGtk::set_click_sensitivity(int value)
{
  click_sensitivity = value;
}

int GlowDrawGtk::export_image(char* filename)
{
  DrawWindGtk* w = &m_wind;
  GdkPixbuf* image = gdk_pixbuf_get_from_drawable(NULL, w->window, 0, 0, 0, 0,
        0, ctx->mw.window_width, ctx->mw.window_height);

  if (!image)
    return 0;

  gboolean sts = gdk_pixbuf_save(image, filename, "png", NULL, "compression",
      "9", NULL);
  if (sts == 0)
    return 0;

  return 1;
}

int GlowDrawGtk::print(char* filename, double x0, double x1, int end)
{
#define ps_cPageHeight 820
#define ps_cPageWidth 535
#define ps_cLeftMargin 100
#define ps_cTopMargin 100

  double x = ps_cLeftMargin;
  double y = ps_cPageHeight - ps_cTopMargin;

  GdkPixbuf* image = gdk_pixbuf_get_from_drawable(
      NULL, ((DrawWindGtk*)w->window)->window, 0, 0, 0, 0, 0, w->window_width, w->window_height);
  if (!image)
    return 0;

  static DrawPs* ps = 0;
  bool new_file = !ps;
  if (!ps) {
    ps = new DrawPs(filename);
    ps->y = y;
  } else
    y = ps->y;

  int width = gdk_pixbuf_get_width(image);
  int height = gdk_pixbuf_get_height(image);
  double scalex = 0.71;
  double scaley = 0.78;

  if (!feq(x0, 0.0) || !feq(x1, 0.0)) {
    double total_width = width / (x1 - x0);

    if (total_width * scalex > ps_cPageWidth - ps_cLeftMargin) {
      x = ps_cPageWidth - total_width * scalex;
      if (x < 50) {
        double scale_factor = (ps_cPageWidth - 50) / (total_width * scalex);
        x = 50;
        scalex = scalex * scale_factor;
        scaley = scaley * scale_factor;
      }
    }
    x += scalex * total_width * x0;
  } else if (width * scalex > ps_cPageWidth - ps_cLeftMargin) {
    x = ps_cPageWidth - width * scalex;
    if (x < 50) {
      double scale_factor = (ps_cPageWidth - 50) / (width * scalex);
      x = 50;
      scalex = scalex * scale_factor;
      scaley = scaley * scale_factor;
    }
  }
  if ((feq(x0, 0.0) && feq(x1, 0.0)) || feq(x1, 1.0))
    ps->y -= scaley * height;

  if (new_file) {
    ps->fp << "%!PS-Adobe-2.0 EPSF-1.2" << endl
           << "%%Creator: Proview Glow" << endl
           << "%%EndComments" << endl
           << endl;
  } else
    ps->fp << "restore" << endl;

  ps->fp << "1.000000 1.000000 scale" << endl
         << "save" << endl
         << scalex * width << " " << scaley * height << " scale" << endl
         << "/oneline " << width << " string def" << endl
         << "/drawimage {" << endl
         << " " << width << " " << height << " 8 [" << width << " 0 0 -"
         << height << " 0 " << height << "]" << endl
         << " { currentfile oneline readhexstring pop }" << endl;
  ps->fp << "false 3" << endl << "colorimage" << endl;

  ps->fp << "} def" << endl
         << x / scalex / width << " " << (y - height * scaley) / scaley / height
         << " translate" << endl
         << "drawimage" << endl;

  ps->fp.flags(
      (ps->fp.flags() & ~ios_base::dec) | ios_base::hex | ios_base::uppercase);
  ps->fp.fill('0');
  unsigned char* rgb = gdk_pixbuf_get_pixels(image);
  int rowstride = gdk_pixbuf_get_rowstride(image);
  int n_channels = gdk_pixbuf_get_n_channels(image);
  int j = 0;
  unsigned char* rgb_row = rgb;
  for (int k = 0; k < height; k++) {
    rgb = rgb_row;
    for (int i = 0; i < width; i++) {
      int red = 255, green = 255, blue = 255;
      if (n_channels < 4 || (rgb + 3)) {
        red = *rgb;
        green = *(rgb + 1);
        blue = *(rgb + 2);
      }
      rgb += n_channels;
      ps->fp.width(2);
      ps->fp << red;
      ps->fp.width(2);
      ps->fp << green;
      ps->fp.width(2);
      ps->fp << blue;
      if (++j >= 20) {
        j = 0;
        ps->fp << endl;
      }
    }
    rgb_row += rowstride;
  }

  if (end) {
    ps->fp << endl << "restore" << endl << "showpage" << endl;
    delete ps;
    ps = 0;
  } else {
    ps->fp.flags(((ps->fp.flags() & ~ios_base::hex) & ~ios_base::uppercase)
        | ios_base::dec);
  }
  g_object_unref(image);

  return 1;
}

GdkPoint* GlowDrawGtk::points_to_gdk_points(glow_sPointX* points, int point_cnt)
{
  GdkPoint* gpoints = (GdkPoint*)malloc(point_cnt * sizeof(GdkPoint));
  for (int i = 0; i < point_cnt; i++) {
    gpoints[i].x = points[i].x;
    gpoints[i].y = points[i].y;
  }
  return gpoints;
}

GdkPoint* GlowDrawGtk::points_to_gdk_points_curve(glow_sPointX* points,
    int point_cnt, int* cnt)
{
  GdkPoint* gpoints = (GdkPoint*)malloc(point_cnt * sizeof(GdkPoint));
  int idx = 0;
  int last_idx = 0;
  for (int i = 0; i < point_cnt; i++) {
    if (idx == 0 && i != point_cnt - 1
        && ((points[i].x < w->subwindow_x && points[i + 1].x < w->subwindow_x)
               || (points[i].x > w->subwindow_x + w->window_width
                      && points[i + 1].x > w->subwindow_x + w->window_width)))

      continue; // Points outside window in x direction excluded
    if (idx != 0
        && !((points[i].x < w->subwindow_x && points[i - 1].x < w->subwindow_x)
               || (points[i].x > w->subwindow_x + w->window_width
                      && points[i - 1].x > w->subwindow_x + w->window_width)))
      last_idx = idx;
    gpoints[idx].x = points[i].x;
    gpoints[idx].y = points[i].y;
    idx++;
  }
  *cnt = idx;
  if (last_idx != 0)
    *cnt = last_idx + 1;
  return gpoints;
}

void GlowDrawGtk::load_font(glow_eFont font_idx, int font_type, int idx)
{
  if (!font[font_idx][font_type][idx])
    font[font_idx][font_type][idx]
        = gdk_font_load(font_names[font_idx][font_type][idx]);
}

int GlowDrawGtk::get_font_type(int gc_type)
{
  if (gc_type == glow_eDrawType_TextHelveticaBold ||
      gc_type == glow_eDrawType_TextHelveticaEraseBold) {
    return glow_eDrawFont_HelveticaBold;
  } else {
    return glow_eDrawFont_Helvetica;
  }
}

int GlowDrawGtk::image_get_width(glow_tImImage image)
{
  return gdk_pixbuf_get_width((GdkPixbuf*)image);
}

int GlowDrawGtk::image_get_height(glow_tImImage image)
{
  return gdk_pixbuf_get_height((GdkPixbuf*)image);
}

int GlowDrawGtk::image_get_rowstride(glow_tImImage image)
{
  return gdk_pixbuf_get_rowstride((GdkPixbuf*)image);
}

unsigned char* GlowDrawGtk::image_get_data(glow_tImImage image)
{
  return gdk_pixbuf_get_pixels((GdkPixbuf*)image);
}

void GlowDrawGtk::image_copy(glow_tImImage orig_image, glow_tImImage* image)
{
  if (*image)
    g_object_unref((GdkPixbuf*)*image);

  *(GdkPixbuf**)image = gdk_pixbuf_copy((GdkPixbuf*)orig_image);
}

void GlowDrawGtk::image_rotate(
    glow_tImImage* image, int to_rotation, int from_rotation)
{
#if GDK_PIXBUF_MAJOR == 2 && GDK_PIXBUF_MINOR < 8
  return;
#else
  int grot;
  int drot = to_rotation - from_rotation;
  drot = int((float(drot) / 360 - floor(float(drot) / 360)) * 360);
  switch (drot) {
  case 90:
    grot = GDK_PIXBUF_ROTATE_CLOCKWISE;
    break;
  case 180:
    grot = GDK_PIXBUF_ROTATE_UPSIDEDOWN;
    break;
  case 270:
    grot = GDK_PIXBUF_ROTATE_COUNTERCLOCKWISE;
    break;
  default:
    return;
  }
  GdkPixbuf* im
      = gdk_pixbuf_rotate_simple((GdkPixbuf*)*image, (GdkPixbufRotation)grot);
  g_object_unref((GdkPixbuf*)*image);
  *image = (glow_tImImage)im;
#endif
}

void GlowDrawGtk::image_flip_vertical(glow_tImImage* image)
{
#if GDK_PIXBUF_MAJOR == 2 && GDK_PIXBUF_MINOR < 8
  return;
#else
  GdkPixbuf* im = gdk_pixbuf_flip((GdkPixbuf*)*image, TRUE);
  g_object_unref((GdkPixbuf*)*image);
  *image = (glow_tImImage)im;
#endif
}

void GlowDrawGtk::image_flip_horizontal(glow_tImImage* image)
{
#if GDK_PIXBUF_MAJOR == 2 && GDK_PIXBUF_MINOR < 8
  return;
#else
  GdkPixbuf* im = gdk_pixbuf_flip((GdkPixbuf*)*image, FALSE);
  g_object_unref((GdkPixbuf*)*image);
  *image = (glow_tImImage)im;
#endif
}

int GlowDrawGtk::image_scale(int width, int height, glow_tImImage orig_im,
    glow_tImImage* im, glow_tImData* im_data, glow_tPixmap* im_pixmap,
    glow_tPixmap* im_mask)
{
  if (width == gdk_pixbuf_get_width((GdkPixbuf*)*im)
      && height == gdk_pixbuf_get_height((GdkPixbuf*)*im))
    return 0;
  if (width == 0 && height == 0) {
    width = gdk_pixbuf_get_width((GdkPixbuf*)*im);
    height = gdk_pixbuf_get_height((GdkPixbuf*)*im);
  }

  if (!orig_im) {
    // Scale from im
    if (!*im)
      return 0;

    GdkPixbuf* im_old = (GdkPixbuf*)*im;
    *im = gdk_pixbuf_scale_simple(
        (GdkPixbuf*)*im, width, height, GDK_INTERP_NEAREST);
    g_object_unref(im_old);
  } else {
    // Scale from orig_im

#if defined PWRE_CONF_RSVG
    if (im_data && *im_data && *im) {
      RsvgHandle* handle;
      const char* imagefile;

      g_object_unref((GdkPixbuf*)*im);

      handle = (RsvgHandle*)*im_data;
      imagefile = rsvg_handle_get_base_uri(handle);
      *im = rsvg_pixbuf_from_file_at_size(imagefile, width, height, NULL);
    } else
#endif
    {
      if (*im)
        g_object_unref((GdkPixbuf*)*im);
      *im = gdk_pixbuf_scale_simple(
          (GdkPixbuf*)orig_im, width, height, GDK_INTERP_NEAREST);
    }
  }
  return 1;
}

int GlowDrawGtk::image_load(char* imagefile, glow_tImImage* orig_im,
    glow_tImImage* im, glow_tImData* im_data)
{
  if (im && *im)
    g_object_unref((GdkPixbuf*)*im);
  if (*orig_im)
    g_object_unref((GdkPixbuf*)*orig_im);
#if defined PWRE_CONF_RSVG
  if (im_data && streq(strchr(imagefile, '.'), ".svg")) {
    RsvgHandle* handle;

    handle = rsvg_handle_new_from_file(imagefile, NULL);
    *im_data = (glow_tImImage*)handle;
    *orig_im = (glow_tImImage*)rsvg_pixbuf_from_file(imagefile, NULL);
  } else
#endif
  {
    *orig_im = (glow_tImImage*)gdk_pixbuf_new_from_file(imagefile, 0);
    if (!*orig_im) {
      if (im)
        *im = 0;
      return 0;
    }
  }
  if (im) {
    *im = (glow_tImImage*)gdk_pixbuf_copy((GdkPixbuf*)*orig_im);
  }
  return 1;
}

void GlowDrawGtk::image_free(glow_tImImage image)
{
  g_object_unref((GdkPixbuf*)image);
}

void GlowDrawGtk::image_pixel_iter(glow_tImImage orig_image,
    glow_tImImage* image, void (*pixel_cb)(void*, unsigned char*),
    void* userdata)
{
  unsigned char *rgb, *rgb_row;
  int rgb_height;
  int rgb_width;
  int rowstride;
  int n_channels;

  if (orig_image) {
    if (image && *image != orig_image)
      g_object_unref((GdkPixbuf*)*image);
    *image = (glow_tImImage*)gdk_pixbuf_copy((GdkPixbuf*)orig_image);
  } else if (!*image)
    return;

  rgb = gdk_pixbuf_get_pixels((GdkPixbuf*)*image);
  rgb_height = gdk_pixbuf_get_height((GdkPixbuf*)*image);
  rgb_width = gdk_pixbuf_get_width((GdkPixbuf*)*image);
  rowstride = gdk_pixbuf_get_rowstride((GdkPixbuf*)*image);
  n_channels = gdk_pixbuf_get_n_channels((GdkPixbuf*)*image);

  rgb_row = rgb;
  for (int j = 0; j < rgb_height; j++) {
    rgb = rgb_row;
    for (int i = 0; i < rgb_width; i++) {
      if (n_channels >= 4 && *(rgb + 3))
        (pixel_cb)(userdata, rgb);
      rgb += n_channels;
    }
    rgb_row += rowstride;
  }
}

void GlowDrawGtk::set_cairo_clip(cairo_t* cr)
{
  DrawWindGtk* w = ((DrawWindGtk*)this->w->window);
  if (w->clip_cnt > 0) {
    cairo_rectangle(cr, w->clip_rectangle[w->clip_cnt - 1].x,
        w->clip_rectangle[w->clip_cnt - 1].y,
        w->clip_rectangle[w->clip_cnt - 1].width,
        w->clip_rectangle[w->clip_cnt - 1].height);
    cairo_clip(cr);
  }
}

void GlowDrawGtk::reset_cairo_clip(cairo_t* cr)
{
  DrawWindGtk* w = ((DrawWindGtk*)this->w->window);
  if (w->clip_cnt > 0) {
    cairo_reset_clip(cr);
  }
}

int GlowDrawGtk::gradient_create_pattern(int x, int y, int w, int h,
    glow_eDrawType d0, glow_eDrawType d1, glow_eDrawType d2,
    glow_eGradient gradient, cairo_pattern_t** pat)
{
  double r0, g0, b0, r1, g1, b1, r2, g2, b2;
  int a = 15;

  GlowColor::rgb_color(d1, &r1, &g1, &b1, get_customcolors());
  GlowColor::rgb_color(d2, &r2, &g2, &b2, get_customcolors());

  double x0 = (w > h) ? (x + w / 2 - h / 2) : (x - (h - w) / a);
  double x1 = (w > h) ? (x + w / 2 + h / 2) : (x + w + (h - w) / a);
  double y0 = (w > h) ? (y - (w - h) / a)     : (y + h / 2 - w / 2);
  double y1 = (w > h) ? (y + h + (w - h) / a) : (y + h / 2 + w / 2);

  switch (gradient) {
  case glow_eGradient_HorizontalUp:
    *pat = cairo_pattern_create_linear(x, y, x, y + h);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    return 1;
  case glow_eGradient_HorizontalDown:
    *pat = cairo_pattern_create_linear(x, y, x, y + h);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r2, g2, b2);
    return 1;
  case glow_eGradient_HorizontalTube1:
    *pat = cairo_pattern_create_linear(x, y, x, y + h);
    cairo_pattern_add_color_stop_rgb(*pat, 0, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.5, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 1, r1, g1, b1);
    return 1;
  case glow_eGradient_HorizontalTube2:
    *pat = cairo_pattern_create_linear(x, y, x, y + h);
    GlowColor::rgb_color(d0, &r0, &g0, &b0, get_customcolors());
    cairo_pattern_add_color_stop_rgb(*pat, 0, r0, g0, b0);
    cairo_pattern_add_color_stop_rgb(*pat, 0.3, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, .95, r1, g1, b1);
    return 1;
  case glow_eGradient_VerticalRight:
    *pat = cairo_pattern_create_linear(x, y, x + w, y);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r2, g2, b2);
    return 1;
  case glow_eGradient_VerticalLeft:
    *pat = cairo_pattern_create_linear(x, y, x + w, y);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    return 1;
  case glow_eGradient_VerticalTube1:
    *pat = cairo_pattern_create_linear(x, y, x + w, y);
    cairo_pattern_add_color_stop_rgb(*pat, 0, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.5, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 1, r1, g1, b1);
    return 1;
  case glow_eGradient_VerticalTube2:
    GlowColor::rgb_color(d0, &r0, &g0, &b0, get_customcolors());
    *pat = cairo_pattern_create_linear(x, y, x + w, y);
    cairo_pattern_add_color_stop_rgb(*pat, 0, r0, g0, b0);
    cairo_pattern_add_color_stop_rgb(*pat, 0.3, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 0.95, r1, g1, b1);
    return 1;
  case glow_eGradient_DiagonalUpperLeft:
    *pat = cairo_pattern_create_linear(x0, y0, x1, y1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    return 1;
  case glow_eGradient_DiagonalLowerLeft:
    *pat = cairo_pattern_create_linear(x0, y1, x1, y0);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    return 1;
  case glow_eGradient_DiagonalUpperRight:
    *pat = cairo_pattern_create_linear(x0, y1, x1, y0);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r2, g2, b2);
    return 1;
  case glow_eGradient_DiagonalLowerRight:
    *pat = cairo_pattern_create_linear(x0, y0, x1, y1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r2, g2, b2);
    return 1;
  case glow_eGradient_DiagonalUpTube:
    *pat = cairo_pattern_create_linear(x0, y0, x1, y1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.5, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    return 1;
  case glow_eGradient_DiagonalDownTube:
    *pat = cairo_pattern_create_linear(x0, y1, x1, y0);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.5, r2, g2, b2);
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    return 1;
  case glow_eGradient_Globe:
    *pat = cairo_pattern_create_radial(x + w / 3, y + h / 3, 0, x + w / 3,
        y + h / 3, sqrt(w * w / 4 + h * h / 4));
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    return 1;
  case glow_eGradient_RadialCenter:
    *pat = cairo_pattern_create_radial(x + w / 2, y + h / 2, 0, x + w / 2,
        y + h / 2, sqrt(w * w / 4 + h * h / 4));
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    return 1;
  case glow_eGradient_RadialUpperLeft:
    *pat = cairo_pattern_create_radial(x, y, 0, x, y, sqrt(w * w + h * h));
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    return 1;
  case glow_eGradient_RadialLowerLeft:
    *pat = cairo_pattern_create_radial(
        x, y + h, 0, x, y + h, sqrt(w * w + h * h));
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    return 1;
  case glow_eGradient_RadialUpperRight:
    *pat = cairo_pattern_create_radial(
        x + w, y, 0, x + w, y, sqrt(w * w + h * h));
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    return 1;
  case glow_eGradient_RadialLowerRight:
    *pat = cairo_pattern_create_radial(
        x + w, y + h, 0, x + w, y + h, sqrt(w * w + h * h));
    cairo_pattern_add_color_stop_rgb(*pat, 0.9, r1, g1, b1);
    cairo_pattern_add_color_stop_rgb(*pat, 0.1, r2, g2, b2);
    return 1;
  default:
    return 0;
  }
}

void GlowDrawGtk::gradient_fill_rect(int x, int y, int width, int h,
    glow_eDrawType d0, glow_eDrawType d1, glow_eDrawType d2,
    glow_eGradient gradient)
{
  cairo_t* cr = gdk_cairo_create(((DrawWindGtk*)w->window)->window);

  set_cairo_clip(cr);

  cairo_pattern_t* pat;
  if (!gradient_create_pattern(x, y, width, h, d0, d1, d2, gradient, &pat))
    return;

  cairo_rectangle(cr, x, y, width, h);
  cairo_set_source(cr, pat);
  cairo_fill(cr);

  reset_cairo_clip(cr);
  cairo_pattern_destroy(pat);
  cairo_destroy(cr);
}

void GlowDrawGtk::gradient_fill_rectrounded(int x, int y, int width,
    int h, int roundamount, glow_eDrawType d0, glow_eDrawType d1,
    glow_eDrawType d2, glow_eGradient gradient)
{
  cairo_t* cr = gdk_cairo_create(((DrawWindGtk*)w->window)->window);

  set_cairo_clip(cr);

  cairo_pattern_t* pat;
  if (!gradient_create_pattern(x, y, width, h, d0, d1, d2, gradient, &pat))
    return;

  if (roundamount >= 0) {
    cairo_move_to(cr, x + roundamount, y);
    cairo_line_to(cr, x + width - roundamount, y);
    cairo_arc(cr, x + width - roundamount, y + roundamount, roundamount,
        -M_PI / 2, 0);
    cairo_line_to(cr, x + width, y + h - roundamount);
    cairo_arc(cr, x + width - roundamount, y + h - roundamount, roundamount, 0,
        M_PI / 2);
    cairo_line_to(cr, x + roundamount, y + h);
    cairo_arc(cr, x + roundamount, y + h - roundamount, roundamount, M_PI / 2,
        M_PI);
    cairo_line_to(cr, x, y + roundamount);
    cairo_arc(cr, x + roundamount, y + roundamount, roundamount, M_PI,
        M_PI * 3 / 2);
  } else
    cairo_rectangle(cr, x, y, width, h);

  cairo_set_source(cr, pat);
  cairo_fill(cr);

  reset_cairo_clip(cr);
  cairo_pattern_destroy(pat);
  cairo_destroy(cr);
}

void GlowDrawGtk::gradient_fill_arc(int x, int y, int width, int h,
    int angle1, int angle2, glow_eDrawType d0, glow_eDrawType d1,
    glow_eDrawType d2, glow_eGradient gradient)
{
  if (angle1 >= 360)
    angle1 = angle1 - angle1 / 360 * 360;
  else if (angle1 < 0)
    angle1 = angle1 + (-angle1 / 360 + 1) * 360;

  cairo_t* cr = gdk_cairo_create(((DrawWindGtk*)w->window)->window);

  set_cairo_clip(cr);

  cairo_pattern_t* pat;
  if (!gradient_create_pattern(x, y, width, h, d0, d1, d2, gradient, &pat))
    return;

  cairo_save(cr);
  cairo_translate(cr, double(x) + double(width) / 2, double(y) + double(h) / 2);
  cairo_scale(cr, double(width) / 2, double(h) / 2);
  if (!(angle2 == 360 || angle2 == 180))
    cairo_move_to(cr, 0, 0);
  cairo_arc(cr, 0, 0, 1, -double(angle1 + angle2) / 180 * M_PI,
      -double(angle1) / 180 * M_PI);
  if (!(angle2 == 360 || angle2 == 180))
    cairo_move_to(cr, 0, 0);
  cairo_restore(cr);
  cairo_set_source(cr, pat);
  cairo_fill(cr);

  reset_cairo_clip(cr);
  cairo_pattern_destroy(pat);
  cairo_destroy(cr);
}

void GlowDrawGtk::gradient_fill_polyline(glow_sPointX* points,
    int point_cnt, glow_eDrawType d0, glow_eDrawType d1, glow_eDrawType d2,
    glow_eGradient gradient)
{
  cairo_t* cr = gdk_cairo_create(((DrawWindGtk*)w->window)->window);

  set_cairo_clip(cr);

  double x0, y0, x1, y1;
  x0 = y0 = 1e37;
  x1 = y1 = 1e-37;
  for (int j = 0; j < point_cnt; j++) {
    if (points[j].x < x0)
      x0 = points[j].x;
    if (points[j].y < y0)
      y0 = points[j].y;
    if (points[j].x > x1)
      x1 = points[j].x;
    if (points[j].y > y1)
      y1 = points[j].y;

    if (j == 0)
      cairo_move_to(cr, points[j].x, points[j].y);
    else
      cairo_line_to(cr, points[j].x, points[j].y);
  }

  cairo_pattern_t* pat;
  if (!gradient_create_pattern((int)x0, (int)y0, (int)(x1 - x0),
          (int)(y1 - y0), d0, d1, d2, gradient, &pat))
    return;
  cairo_set_source(cr, pat);
  cairo_fill(cr);

  reset_cairo_clip(cr);
  cairo_pattern_destroy(pat);
  cairo_destroy(cr);
}

void GlowDrawGtk::log_event(GdkEvent* event)
{
  if (ctx->eventlog_callback) {
    GdkEvent e;

    memcpy(&e, event, sizeof(e));

    switch (e.type) {
    case GDK_KEY_PRESS:
      *(gchar*)&e.key.string = e.key.string[0];
      break;
    case GDK_MOTION_NOTIFY: {
      int x, y;

      if (e.motion.is_hint) {
        gdk_window_get_pointer(e.any.window, &x, &y, NULL);
        e.button.x = x;
        e.button.y = y;
        e.motion.is_hint = 0;
      }
      break;
    }
    case GDK_EXPOSE: {
      int width, height;

      get_window_size(ctx->mw.window, &width, &height);
      if (e.expose.area.x == 0 && e.expose.area.y == 0) {
        e.expose.area.x = -999;
        e.expose.area.y = -999;
        e.expose.area.width = width;
        e.expose.area.height = height;
      }
      break;
    }
    default:;
    }

    (ctx->eventlog_callback)(ctx, &e, sizeof(e));
  }
}

void GlowDrawGtk::event_exec(void* event, unsigned int size)
{
  GdkEvent e;

  if (size != sizeof(e))
    return;

  memcpy(&e, event, size);
  e.any.window = m_wind.window;

  switch (e.type) {
  case GDK_KEY_PRESS: {
    gchar* p = (gchar*)malloc(1);

    *p = *(gchar*)&e.key.string;
    e.key.string = p;
    break;
  }
  case GDK_EXPOSE: {
    if (e.expose.area.x == -999 && e.expose.area.y == -999) {
      set_window_size(ctx->mw.window, e.expose.area.width, e.expose.area.height);
      e.expose.area.x = 0;
      e.expose.area.y = 0;
    }
    break;
  }
  default:;
  }

  event_handler(e);

  switch (e.type) {
  case GDK_KEY_PRESS: {
    free(e.key.string);
    break;
  }
  default:;
  }
}

int GlowDrawGtk::open_color_selection(double* r, double* g, double* b)
{
  int sts;
  GdkColor color;
  GtkWidget* w = gtk_color_selection_dialog_new("Color selection");
  GtkWidget* csel = gtk_color_selection_dialog_get_color_selection(
      GTK_COLOR_SELECTION_DIALOG(w));
  color.red = int(*r * 0xffff);
  color.green = int(*g * 0xffff);
  color.blue = int(*b * 0xffff);
  gtk_color_selection_set_current_color(GTK_COLOR_SELECTION(csel), &color);
  gint response = gtk_dialog_run(GTK_DIALOG(w));
  if (response == GTK_RESPONSE_OK) {
    gtk_color_selection_get_current_color(GTK_COLOR_SELECTION(csel), &color);
    *r = (double)color.red / 0xffff;
    *g = (double)color.green / 0xffff;
    *b = (double)color.blue / 0xffff;
    sts = 1;
  } else
    sts = 0;
  gtk_widget_destroy(w);
  return sts;
}

void GlowDrawGtk::update_color(glow_eDrawType color)
{
  if (!GlowColor::is_custom(color)) {
    for (int j = 0; j < DRAW_TYPE_SIZE; j++) {
      if (gcs[color][j]) {
        gdk_gc_unref(gcs[color][j]);
        gcs[color][j] = 0;
      }
    }
  } else if (get_customcolors()) {
    // A customcolor
    for (int i = 0; i < 4; i++) {
      for (int j = 0; j < DRAW_TYPE_SIZE; j++) {
        if (get_customcolors()->gcs[color - glow_eDrawType_CustomColor1 + i][j]) {
          gdk_gc_unref(get_customcolors()
                           ->gcs[color - glow_eDrawType_CustomColor1 + i][j]);
          get_customcolors()->gcs[color - glow_eDrawType_CustomColor1 + i][j] = 0;
        }
      }
    }
  }
}

void GlowDrawGtk::push_customcolors(GlowCustomColors* cc)
{
  if (customcolors_cnt > CUSTOMCOLORS_STACK_SIZE) {
    printf("** Max number custom colors exceede\n");
    return;
  }

  for (int i = customcolors_cnt; i > 0; i--)
    customcolors[i] = customcolors[i - 1];
  customcolors[0] = (GlowCustomColorsGtk*)cc;
  customcolors_cnt++;
}

void GlowDrawGtk::set_customcolors(GlowCustomColors* cc)
{
  customcolors[0] = (GlowCustomColorsGtk*)cc;
  customcolors_cnt = 1;
}

void GlowDrawGtk::pop_customcolors()
{
  if (customcolors_cnt <= 0) {
    printf("** Customcolor stack disorder\n");
    return;
  }

  for (int i = 0; i < customcolors_cnt - 1; i++)
  {
    customcolors[i] = customcolors[i + 1];
    customcolors[i + 1] = NULL;
  }
  customcolors_cnt--;
}

GlowCustomColorsGtk* GlowDrawGtk::get_customcolors()
{
  if (customcolors_cnt == 0)
    return NULL;
  return customcolors[0];
}

GlowCustomColors* GlowDrawGtk::create_customcolors()
{
  return new GlowCustomColorsGtk();
}

void GlowDrawGtk::reset_customcolors(GlowCustomColors* cc)
{
  for (int i = 0; i < cc->colors_size; i++) {
    for (int j = 0; j < DRAW_TYPE_SIZE; j++) {
      if (((GlowCustomColorsGtk*)cc)->gcs[i][j])
        gdk_gc_unref(((GlowCustomColorsGtk*)cc)->gcs[i][j]);
    }
  }
  memset(((GlowCustomColorsGtk*)cc)->gcs, 0,
      sizeof(((GlowCustomColorsGtk*)cc)->gcs));

  for (int i = 0;
       i < glow_eDrawType_CustomColor__ - glow_eDrawType_CustomColor1; i++) {
    if (((GlowCustomColorsGtk*)cc)->color_vect[i].pixel != 0)
      gdk_colormap_free_colors(
          colormap, &((GlowCustomColorsGtk*)cc)->color_vect[i], 1);
  }
  memset(((GlowCustomColorsGtk*)cc)->color_vect, 0,
      sizeof(((GlowCustomColorsGtk*)cc)->color_vect));
}
